from django.conf import settings
from django.utils.six import moves


DJANGO_SCIM_BASE_LOCATION_GETTER = 'django_scim.utils.default_base_scim_location_getter'

DEFAULT_USER_ADAPTER_STRING = 'django_scim.adapters.SCIMUser'

DEFAULT_GROUP_MODEL_STRING = 'django.contrib.auth.models.Group'
DEFAULT_GROUP_ADAPTER_STRING = 'django_scim.adapters.SCIMGroup'

DOCUMENTATION_URI = getattr(settings, 'DJANGO_SCIM_SP_DOC_URL', None)

SCHEMA_URI_ERROR = 'urn:ietf:params:scim:api:messages:2.0:Error'




from django.conf import settings
from django.utils.module_loading import import_string


class SCIMExampleAuth(object):
    """
    Example auth class. Given a request and its args/kwargs,
    this class can authenticate and authorize the request.
    """

    type = 'httpbasic'
    name = 'Http Basic'
    description = 'Authentication scheme using the HTTP Basic Standard'
    spec_uri = 'http://www.rfc-editor.org/info/rfc2617'
    documentation_uri = 'http://exmaple.com/docs.html'

    def __init__(self, request, *args, **kwargs):
        self.request = request
        self.args = args
        self.kwargs = kwargs

    def raise_auth_exceptions(self):
        """
        To reject a request for whatever reason, raise an exception.
        To allow a request, do nothing.
        """
        # raise errors in here
        # exceptions need a 'status' attr and must be castable to a string.
        pass

    @classmethod
    def scheme_dict(cls):
        return {
            'type': self.type,
            'name': self.name,
            'description': self.description,
            'specUri': self.spec_uri,
            'documentationUri': self.documentation_uri,
        }


class SCIMAuthBackendCollection(object):

    _backends = None

    @classmethod
    def backends(cls):
        if cls._backends is None:
            cls_strings = getattr(settings, 'DJANGO_SCIM_AUTH_BACKENDS', [])
            cls._backends = []
            for string in cls_strings:
               cls._backends.append(import_string(string))
        return cls._backends


class SCIMRequest(object):
    def __init__(self, request, *args, **kwargs):
        self.request = request
        self.args = args
        self.kwargs = kwargs

    def raise_auth_execptions(self):
        for backend in SCIMAuthBackendCollection.backends():
            b = backend(self.request, *self.request.args, **self.request.kwargs)
            b.raise_auth_exceptions()




from django.conf import settings
from django.utils.module_loading import import_string

from .constants import DEFAULT_USER_ADAPTER_STRING
from .constants import DEFAULT_GROUP_MODEL_STRING
from .constants import DEFAULT_GROUP_ADAPTER_STRING


def get_user_adapter():
    model_str = getattr(settings, 'DJANGO_SCIM_USER_ADAPTER', DEFAULT_USER_ADAPTER_STRING)
    return import_string(model_str)


def get_group_model():
    model_str = getattr(settings, 'DJANGO_SCIM_GROUP_MODEL', DEFAULT_GROUP_MODEL_STRING)
    return import_string(model_str)


def get_group_adapter():
    model_str = getattr(settings, 'DJANGO_SCIM_GROUP_ADAPTER', DEFAULT_GROUP_ADAPTER_STRING)
    return import_string(model_str)


def get_base_scim_location_getter():
    """
    Return a function that will, when called, return the base
    location of scim app.
    """
    import_str = getattr(
        settings,
        'DJANGO_SCIM_BASE_LOCATION_GETTER',
        DEFAULT_BASE_LOCATION_GETTER_STRING
    )
    return import_string(import_str)


def default_base_scim_location_getter():
    base_scim_location_parts = (
        getattr(settings, 'DJANGO_SCIM_SCHEME', 'https'),
        settings.DJANGO_SCIM_NETLOC,
        '',  # path
        '',  # params
        '',  # query
        ''   # fragment
    )

    base_scim_location = moves.urllib.parse.urlunparse(base_scim_location_parts)

    return base_scim_location




import dateutil.parser
import itertools
import re

from plyplus import Grammar, STransformer, PlyplusException
from django.contrib.auth import get_user_model


STRING_REPLACEMENT_RE_PAT = re.compile(r'\%\([^).]+\)s', re.MULTILINE)


grammar = Grammar("""
  start: logical_or;

  ?logical_or: logical_or op_or logical_and | logical_and;
  ?logical_and: logical_and op_and expr | expr;

  ?expr: un_string_expr | un_expr | bin_expr | '\(' logical_or '\)';

  ?bin_expr: (bin_string_expr | bin_passwd_expr | bin_pk_expr | bin_date_expr |
              bin_bool_expr);

  un_string_expr: (string_field | password) pr;
  un_expr: (date_field | bool_field) pr;
  bin_string_expr: string_field string_op t_string;
  bin_passwd_expr: password eq t_string;
  bin_pk_expr: pk eq t_string;
  bin_date_expr: date_field date_op t_date;
  bin_bool_expr: bool_field eq t_bool;

  ?string_op: eq | co | sw;
  ?date_op: eq | gt | ge | lt | le;

  ?op_or: '(?i)or';
  ?op_and: '(?i)and';
  pr: '(?i)pr';
  eq: '(?i)eq';
  co: '(?i)co';
  sw: '(?i)sw';
  gt: '(?i)gt';
  ge: '(?i)ge';
  lt: '(?i)lt';
  le: '(?i)le';

  ?string_field: username | first_name | last_name | email;
  ?date_field: date_joined;
  ?bool_field: is_active;

  username: '(?i)userName';
  first_name: '(?i)name\.givenName' | '(?i)givenName';
  last_name: '(?i)name\.familyName' | '(?i)familyName';
  password: '(?i)password';
  pk: '(?i)id';
  date_joined: '(?i)meta\.created' | '(?i)created';
  email: '(?i)emails\.value' | '(?i)emails';
  is_active: '(?i)active';

  t_string: '"(?:[^"\\\\]|\\\\.)*"' ;
  t_date: '"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}[Zz]?"';
  t_bool: 'false' | 'true';

  SPACES: '[ ]+' (%ignore);
  """)


class SCIMUserFilterTransformer(STransformer):
    """Transforms a PlyPlus parse tree into a tuple containing a raw SQL query
    and a dict with query parameters to go with the query."""

    # data types:
    t_string = lambda self, exp: exp.tail[0][1:-1]
    t_date = lambda self, exp: dateutil.parser.parse(exp.tail[0][1:-1])
    t_bool = lambda self, exp: exp.tail[0] == u'true'

    # operators
    op_or = lambda self, exp: u'OR'
    op_and = lambda self, exp: u'AND'
    gt = ge = lt = le = pr = eq = co = sw = lambda self, ex: ex.tail[0].lower()

    # fully qualified column names:
    pk = lambda *args: u'u.id'
    username = lambda *args: u'u.username'
    password = lambda *args: u'u.password'
    first_name = lambda *args: u'u.first_name'
    last_name = lambda *args: u'u.last_name'
    email = lambda *args: u'u.email'
    date_joined = lambda *args: u'u.date_joined'
    is_active = lambda *args: u'u.is_active'

    @property
    def auth_user_db_table(self):
        return get_user_model()._meta.db_table

    # expressions:
    def logical_or(self, exp):
        # We're not doing a simple 'OR', as that doesn't scale when the two
        # conditions operate on different tables and rely on indices.
        # This is because Postgres' query planner cannot leverage indices from
        # different tables. Instead, we'll use a UNION.
        #
        # http://grokbase.com/t/postgresql/pgsql-general/034qrq6me0/left-join-not-using-index
        return u"""
            u.id IN (
                WITH users AS (
                    SELECT DISTINCT u.id
                    FROM {auth_user_db_table} u
                    {join}
                    WHERE
                    {operand1}

                    UNION

                    SELECT DISTINCT u.id
                    FROM {auth_user_db_table} u
                    {join}
                    WHERE
                    {operand2}
                )
                SELECT DISTINCT id FROM users
            )
        """.format(join=self.join(),
                   auth_user_db_table=self.auth_user_db_table,
                   operand1=exp.tail[0],
                   operand2=exp.tail[1])

    def logical_and(self, exp):
        op1, op2 = exp.tail
        if isinstance(op1, self.PasswordExpression):
            params = {'fragment': op2, 'password': op1}
        elif isinstance(op2, self.PasswordExpression):
            params = {'fragment': op1, 'password': op2}
        else:
            return u'(%s AND %s)' % (op1, op2)

        # When expensive CRYPT functions are involved we can't rely on
        # Postgres' query planner to chose the optimal order in which to
        # evaluate the conditions.
        # If a password condition is AND'ed against something else, we
        # explicitly write SQL that will force Postgres to compute the password
        # hashes on the *result* of the other condition, to minimize the number
        # of CRYPT calculations.
        return u"""
            u.id IN
            (
                WITH users AS (
                    SELECT DISTINCT u.id, u.password
                    FROM {auth_user_db_table} u
                    {join}
                    WHERE {fragment}
                )
                SELECT DISTINCT users.id
                FROM users
                WHERE {password}
            )""".format(join=self.join(),
                        auth_user_db_table=self.auth_user_db_table,
                        **params)

    __default__ = lambda self, exp: exp.tail[0]

    def __init__(self):
        self._seq = itertools.count(0, step=1)
        self._params = {}

    def _push_param(self, value):
        name = str(next(self._seq))
        self._params[name] = value
        return name

    def join(self):
        """Returns join expressions. E.g.

            JOIN bb_userprofile p ON p.user_id = u.id
        """
        return ''

    def start(self, exp):
        return u"""
            SELECT DISTINCT u.*
            FROM {auth_user_db_table} u
                {join}
            WHERE {fragment}
            ORDER BY u.id ASC
            """.format(join=self.join(),
                       auth_user_db_table=self.auth_user_db_table,
                       fragment=exp.tail[0]), self._params

    def un_expr(self, exp):
        return u'%s IS NOT NULL' % exp.tail[0]

    def un_string_expr(self, exp):
        # Django uses empty strings instead of NULL in VARCHARs:
        return u"(%s IS NOT NULL AND %s != '')" % (exp.tail[0], exp.tail[0])

    def bin_string_expr(self, exp):
        field, op, literal = exp.tail
        if op == u'eq':
            return u'UPPER(%s) = UPPER(%%(%s)s)' % (field, self._push_param(literal))
        elif op == u'sw':
            literal += u'%'
        elif op == u'co':
            literal = u'%' + literal + u'%'
        return u'%s ILIKE %%(%s)s' % (field, self._push_param(literal))

    def bin_date_expr(self, exp):
        field, op, literal = exp.tail
        op = {u'gt': u'>', u'ge': u'>=', u'lt': u'<', u'le': u'<='}[op]
        return u'%s %s %%(%s)s' % (field, op, self._push_param(literal))

    def bin_bool_expr(self, exp):
        return u'%s = %%(%s)s' % (exp.tail[0], self._push_param(exp.tail[2]))

    def bin_pk_expr(self, exp):
        field, op, value = exp.tail
        return u'%s = %%(%s)s' % (field, self._push_param(int(value)))

    def bin_passwd_expr(self, exp):
        pname = self._push_param(exp.tail[2])
        return self.PasswordExpression(u"""
            (
              CHAR_LENGTH(password) >= 51
              AND
              (
                -- Check for SHA1
                SUBSTRING(password FROM 12) = ENCODE(DIGEST(SUBSTRING(
                    password FROM 6 FOR 5)||%%(%s)s, 'sha1'), 'hex')
                OR
                -- Check for BCrypt
                SUBSTRING(password FROM 8) = CRYPT(
                  %%(%s)s, SUBSTRING(password FROM 8))
              )
            )""" % (pname, pname))

    @classmethod
    def condition_sql_and_params(cls, sql, params):
        # replace %(id)s with %s and update params accordingly
        replacements = STRING_REPLACEMENT_RE_PAT.findall(sql)
        new_sql = STRING_REPLACEMENT_RE_PAT.sub('%s', sql, count=len(replacements))

        new_params = []
        for replacement in replacements:
            # strip '%(' adn ')s' from replacement arg
            replacement = replacement[2:-2]
            new_params.append(params.get(replacement))

        return new_sql, new_params

    @classmethod
    def search(cls, query):
        """Takes a SCIM 1.1 filter query and returns a Django `QuerySet` that
        contains zero or more user model instances.

        :param unicode query: a `unicode` query string.
        """
        try:
            sql, params = cls().transform(grammar.parse(query))
            sql, params = cls.condition_sql_and_params(sql, params)
        except PlyplusException as e:
            raise ValueError(e)
        else:
            return get_user_model().objects.raw(sql, params)

    class PasswordExpression(object):
        def __init__(self, sql):
            assert isinstance(sql, unicode)
            self.sql = sql

        def __str__(self):
            return self.sql.encode('utf-8')

        def __unicode__(self):
            return self.sql




import json

from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django import db
from django.db import transaction
from django.http import HttpResponse
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View
from django.utils import six
from django.utils.six.moves.urllib.parse import urljoin

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse

from .auth import SCIMRequest
from .filter import SCIMUserFilterTransformer
from .exceptions import SCIMException
from .exceptions import NotFound
from .exceptions import BadRequest
from .exceptions import IntegrityError
from .models import SCIMServiceProviderConfig
from .schemas import ALL as ALL_SCHEMAS
from .utils import get_group_adapter
from .utils import get_group_model
from .utils import get_user_adapter
from .utils import get_base_scim_location_getter


SCIM_CONTENT_TYPE = 'application/scim+json'

SCHEMA_URI_SERACH_REQUEST = 'urn:ietf:params:scim:api:messages:2.0:SearchRequest'


class SCIMView(View):

    implemented = True

    @csrf_exempt
    def dispatch(self, request, *args, **kwargs):
        if not self.implemented:
            return self.status_501(request, *args, **kwargs)

        try:
            self.auth_request(request, *args, **kwargs)
            return super(SCIMView, self).dispatch(request, *args, **kwargs)
        except Exception as e:
            if not isinstance(e, SCIMException):
                e = SCIMException(six.text_type(e))

            content = json.dumps(e.to_dict(), encoding='utf-8')
            return HttpResponse(content=content,
                                content_type=SCIM_CONTENT_TYPE,
                                status=e.status)

    def status_501(self, request, *args, **kwargs):
        """
        A service provider that does NOT support a feature SHOULD
        respond with HTTP status code 501 (Not Implemented).
        """
        return HttpResponse(content_type=SCIM_CONTENT_TYPE,
                            status=501)

    def auth_request(self, request, *args, **kwargs):
        SCIMRequest(request, *args, **kwargs).raise_auth_execptions()


class FilterMixin(object):

    parser = None
    scim_adapter = None

    def _page(self, request):
        try:
            start = request.GET.get('startIndex', 1)
            if start is not None:
                start = int(start)
                if start < 1:
                    raise BadRequest('Invalid startIndex (must be >= 1)')

            count = request.GET.get('count', 50)
            if count is not None:
                count = int(count)

            return start, count

        except ValueError as e:
            raise BadRequest('Invalid pagination values: ' + str(e))

    def _search(self, query, start, count):
        qs = self.parser.search(query)
        return self._build_response(qs, start, count)

    def _build_response(self, qs, start, count):
        try:
            total_count = sum(1 for _ in qs)
            qs = qs[start-1:(start-1) + count]
            resources = [self.scim_adapter(u).to_dict() for u in qs]
            doc = {
                'schemas': ['urn:ietf:params:scim:api:messages:2.0:ListResponse'],
                'totalResults': total_count,
                'itemsPerPage': count,
                'startIndex': start,
                'Resources': resources,
            }
        except ValueError as e:
            raise BadRequest(six.text_type(e))
        else:
            content = json.dumps(doc, encoding='utf-8')
            return HttpResponse(content=content,
                                content_type=SCIM_CONTENT_TYPE)


class SearchView(FilterMixin, SCIMView):
    http_method_names = ['post']

    scim_adapter = None

    def post(self, request):
        body = json.loads(request.body or '{}')
        if body.get('schemas') != [SCHEMA_URI_SERACH_REQUEST]:
            raise BadRequest('Invalid schema uri. Must be SearchRequest.')

        query = body.get('filter', request.GET.get('filter'))

        if not query:
            raise BadRequest('No filter query specified')
        else:
            response = self._search(query, *self._page(request))
            path = reverse(self.scim_adapter.url_name)
            url = urljoin(get_base_scim_location_getter()(), path).rstrip('/')
            response['Location'] = url + '/.search'
            return response


class GetView(object):
    def get(self, request, uuid=None):
        if uuid:
            return self.get_single(request, uuid)

        return self.get_many(request)

    def get_single(self, request, uuid):
        try:
            scim_obj = self.scim_adapter(self.model_cls.objects.get(id=uuid))
        except ObjectDoesNotExist as _e:
            raise NotFound(uuid)
        else:
            content = json.dumps(scim_obj.to_dict(), encoding='utf-8')
            response = HttpResponse(content=content,
                                    content_type=SCIM_CONTENT_TYPE)
            response['Location'] = scim_obj.location
            return response

    def get_many(self, request):
        query = request.GET.get('filter')
        if query:
            return self._search(query, *self._page(request))

        qs = self.model_cls.objects.all().order_by('id')
        return self._build_response(qs, *self._page(request))


class DeleteView(object):
    def delete(self, request, uuid):
        obj_qs = self.model_cls.objects.filter(id=uuid)

        if obj_qs.exists():
            obj_qs.delete()
            return HttpResponse(status=204)
        else:
            raise NotFound(uuid)


class PostView(object):
    def post(self, request, **kwargs):
        obj = self.model_cls()
        scim_obj = self.scim_adapter(obj)

        body = json.loads(request.body)

        scim_obj.from_dict(body)
        try:
            scim_obj.save()
        except db.utils.IntegrityError as e:
            raise IntegrityError(str(e))

        content = json.dumps(scim_obj.to_dict(), encoding='utf-8')
        response = HttpResponse(content=content,
                                content_type=SCIM_CONTENT_TYPE,
                                status=201)
        response['Location'] = scim_obj.location
        return response


class PutView(object):
    def put(self, request, uuid):
        try:
            obj = self.model_cls.objects.get(id=uuid)
        except ObjectDoesNotExist:
            raise NotFound(uuid)

        scim_obj = self.scim_adapter(obj)

        body = json.loads(request.body)

        scim_obj.from_dict(body)
        scim_obj.save()

        content = json.dumps(scim_obj.to_dict(), encoding='utf-8')
        response = HttpResponse(content=content,
                                content_type=SCIM_CONTENT_TYPE)
        response['Location'] = scim_obj.location
        return response


class PatchView(object):
    def patch(self, request, uuid):

        try:
            obj = self.model_cls.objects.get(id=uuid)
        except ObjectDoesNotExist:
            raise NotFound(uuid)

        scim_obj = self.scim_adapter(obj)
        body = json.loads(request.body)

        operations = body.get('Operations')

        with transaction.atomic():
            scim_obj.handle_operations(operations)

        content = json.dumps(scim_obj.to_dict(), encoding='utf-8')
        response = HttpResponse(content=content,
                                content_type=SCIM_CONTENT_TYPE)
        response['Location'] = scim_obj.location
        return response


class UsersView(FilterMixin, GetView, PostView, PutView, PatchView, DeleteView, SCIMView):

    http_method_names = ['get', 'post', 'put', 'patch', 'delete']

    scim_adapter = get_user_adapter()
    model_cls = get_user_model()
    parser = SCIMUserFilterTransformer


class GroupsView(FilterMixin, GetView, PostView, PutView, PatchView, DeleteView, SCIMView):

    http_method_names = ['get', 'post', 'put', 'patch', 'delete']

    scim_adapter = get_group_adapter()
    model_cls = get_group_model()
    parser = None


class ServiceProviderConfigView(SCIMView):
    http_method_names = ['get']

    def get(self, request):
        config = SCIMServiceProviderConfig()
        content = json.dumps(config.to_dict(), encoding='utf-8')
        return HttpResponse(content=content,
                            content_type=SCIM_CONTENT_TYPE)


class ResourceTypesView(SCIMView):

    http_method_names = ['get']

    @property
    def type_dict_by_type_id(self):
        type_adapters = get_user_adapter(), get_group_adapter()
        type_dicts = [m.resource_type_dict() for m in type_adapters]
        return {d['id']: d for d in type_dicts}

    def get(self, request, uuid=None, *args, **kwargs):
        if uuid:
            doc = self.type_dict_by_type_id.get(uuid)
            if not doc:
                return HttpResponse(content_type=SCIM_CONTENT_TYPE,
                                    status=404)
        else:
            types = list(sorted(self.type_dict_by_type_id.values()))
            doc = {
                'schemas': ['urn:ietf:params:scim:api:messages:2.0:ListResponse'],
                'Resources': types,
            }

        return HttpResponse(content=json.dumps(doc, encoding='utf-8'),
                            content_type=SCIM_CONTENT_TYPE)


class SchemasView(SCIMView):

    http_method_names = ['get']

    schemas_by_uri = {s['id']: s for s in ALL_SCHEMAS}

    def get(self, request, uuid=None, *args, **kwargs):
        if uuid:
            doc = self.schemas_by_uri.get(uuid)
            if not doc:
                return HttpResponse(content_type=SCIM_CONTENT_TYPE,
                                    status=404)
        else:
            schemas = list(sorted(self.schemas_by_uri.values()))
            doc = {
                'schemas': ['urn:ietf:params:scim:api:messages:2.0:ListResponse'],
                'Resources': schemas,
            }



        content = json.dumps(doc, encoding='utf-8')
        return HttpResponse(content=content,
                            content_type=SCIM_CONTENT_TYPE)

